from typing import TYPE_CHECKING, List, Dict

from pandas import DataFrame, MultiIndex

from ibm_watson_machine_learning.utils.autoai.enums import (
    PredictionType, RegressionAlgorithms, ClassificationAlgorithms)
from ibm_watson_machine_learning.utils.autoai.errors import PipelineNotLoaded, FitNeeded, InvalidPredictionType
from ibm_watson_machine_learning.utils.autoai.progress_bar import ProgressBar
from ibm_watson_machine_learning.utils.autoai.utils import (
    fetch_pipelines, is_ipython, RunDetailsGenerator, check_dependencies_versions, try_import_autoai_libs,
    create_summary)
from ibm_watson_machine_learning.utils.autoai.watson_studio import change_client_get_headers
from .base_engine import BaseEngine

if TYPE_CHECKING:
    from ibm_watson_machine_learning.workspace import WorkSpace
    from ibm_watson_machine_learning.helpers.connections import DataConnection
    from sklearn.pipeline import Pipeline

__all__ = [
    "WMLEngine"
]


class WMLEngine(BaseEngine):
    """
    WML Engine provides unified API to work with AutoAI Pipelines trainings on WML.

    Parameters
    ----------
    workspace: WorkSpace, required
        WorkSpace object with wml client and all needed parameters.
    """

    def __init__(self, workspace: 'WorkSpace') -> None:
        self.workspace = workspace
        self._wml_client = workspace.wml_client
        self._auto_pipelines_parameters = None
        self._wml_pipeline_metadata = None
        self._wml_training_metadata = None
        self._wml_stored_pipeline_details = None
        self._current_run_id = None

    def _initialize_wml_pipeline_metadata(self) -> None:
        """
        Initialization of WML Pipeline Document (WML client Meta Parameter) with provided parameters and default ones.
        """

        # note: if user does not specify algorithms to use, use all available
        if self._auto_pipelines_parameters.get('daub_include_only_estimators') is None:
            if self._auto_pipelines_parameters.get('prediction_type') == PredictionType.REGRESSION:
                self._auto_pipelines_parameters['daub_include_only_estimators'] = [
                    algorithm[1].value for algorithm in RegressionAlgorithms.__members__.items()]
            else:
                self._auto_pipelines_parameters['daub_include_only_estimators'] = [
                    algorithm[1].value for algorithm in ClassificationAlgorithms.__members__.items()]
        # --- end note
        else:
            # note: transform values from Enum
            try:
                self._auto_pipelines_parameters['daub_include_only_estimators'] = [
                    algorithm.value for algorithm in self._auto_pipelines_parameters.get('daub_include_only_estimators')
                ]
            # note: if user pass strings instead of enums
            except AttributeError:
                pass

        self._wml_pipeline_metadata = {
            self._wml_client.pipelines.ConfigurationMetaNames.NAME:
                self._auto_pipelines_parameters.get('name', 'Default name.'),
            self._wml_client.pipelines.ConfigurationMetaNames.DESCRIPTION:
                self._auto_pipelines_parameters.get('desc', 'Default description'),
            self._wml_client.pipelines.ConfigurationMetaNames.DOCUMENT: {
                'doc_type': 'pipeline',
                'version': '2.0',
                'pipelines': [{
                    'id': 'autoai',
                    'runtime_ref': 'hybrid',
                    'nodes': [{
                        'id': 'automl',
                        'type': 'execution_node',
                        'parameters': {
                            'stage_flag': True,
                            'output_logs': True,
                            'input_file_separator': self._auto_pipelines_parameters.get('csv_separator'),
                            'optimization': {
                                'learning_type': self._auto_pipelines_parameters.get('prediction_type'),
                                'label': self._auto_pipelines_parameters.get('prediction_column'),
                                'max_num_daub_ensembles': self._auto_pipelines_parameters.get('max_num_daub_ensembles'),
                                "daub_include_only_estimators":
                                    self._auto_pipelines_parameters.get('daub_include_only_estimators'),
                                'scorer_for_ranking': self._auto_pipelines_parameters.get('scoring'),
                                'compute_pipeline_notebooks_flag': self._auto_pipelines_parameters.get('notebooks',
                                                                                                       False),
                                'run_cognito_flag': True,
                                'holdout_param': self._auto_pipelines_parameters.get('test_size')
                            }
                        },
                        'runtime_ref': 'autoai',
                        'op': 'kube'
                    }]
                }],
                'runtimes': [{
                    'id': 'autoai',
                    'name': 'auto_ai.kb',
                    'app_data': {
                        'wml_data': {
                            'hardware_spec': {}
                        }
                    }
                }],
                'primary_pipeline': 'autoai'
            }
        }

        # note: in new v4 api we have missing hw_spec name
        t_size = self._auto_pipelines_parameters['t_shirt_size']
        if len(t_size) == 1:
            self._wml_pipeline_metadata[self._wml_client.pipelines.ConfigurationMetaNames.DOCUMENT]['runtimes'][0][
                'app_data']['wml_data']['hardware_spec']['name'] = t_size.upper()

        else:
            self._wml_pipeline_metadata[self._wml_client.pipelines.ConfigurationMetaNames.DOCUMENT]['runtimes'][0][
                'app_data']['wml_data']['hardware_spec']['id'] = t_size
        # --- end note

        if self._auto_pipelines_parameters.get('train_sample_rows_test_size') is not None:
            self._wml_pipeline_metadata[self._wml_client.pipelines.ConfigurationMetaNames.DOCUMENT][
                'pipelines'][0]['nodes'][0]['parameters']['optimization'][
                'train_sample_rows_test_size'] = self._auto_pipelines_parameters['train_sample_rows_test_size']

        if self._auto_pipelines_parameters.get('t_shirt_size') == 's':
            self._wml_pipeline_metadata[self._wml_client.pipelines.ConfigurationMetaNames.DOCUMENT][
                'pipelines'][0]['nodes'][0]['parameters']['optimization'][
                'daub_adaptive_subsampling_max_mem_usage'] = 6e9

        elif self._auto_pipelines_parameters.get('t_shirt_size') == 'm':
            self._wml_pipeline_metadata[self._wml_client.pipelines.ConfigurationMetaNames.DOCUMENT][
                'pipelines'][0]['nodes'][0]['parameters']['optimization'][
                'daub_adaptive_subsampling_max_mem_usage'] = 9e9

        elif self._auto_pipelines_parameters.get('t_shirt_size') == 'l':
            self._wml_pipeline_metadata[self._wml_client.pipelines.ConfigurationMetaNames.DOCUMENT][
                'pipelines'][0]['nodes'][0]['parameters']['optimization'][
                'daub_adaptive_subsampling_max_mem_usage'] = 15e9

        elif self._auto_pipelines_parameters.get('t_shirt_size') == 'xl':
            self._wml_pipeline_metadata[self._wml_client.pipelines.ConfigurationMetaNames.DOCUMENT][
                'pipelines'][0]['nodes'][0]['parameters']['optimization'][
                'daub_adaptive_subsampling_max_mem_usage'] = 25e9

        # note: only pass positive label when scoring is binary
        if (self._auto_pipelines_parameters.get('positive_label')
                and self._auto_pipelines_parameters.get('prediction_type') == PredictionType.BINARY):
            self._wml_pipeline_metadata[self._wml_client.pipelines.ConfigurationMetaNames.DOCUMENT][
                'pipelines'][0]['nodes'][0]['parameters']['optimization'][
                'positive_label'] = self._auto_pipelines_parameters.get('positive_label')
        # --- end note

        # note: only pass excel_sheet when it is different than 0
        if self._auto_pipelines_parameters.get('excel_sheet') != 0:
            self._wml_pipeline_metadata[self._wml_client.pipelines.ConfigurationMetaNames.DOCUMENT][
                'pipelines'][0]['nodes'][0]['parameters'][
                'excel_sheet'] = self._auto_pipelines_parameters.get('excel_sheet')
        # --- end note

        # note: OBM part, only if user specified OBM as a preprocess method
        if self._auto_pipelines_parameters['data_join_graph']:
            # note: remove kb params for OBM only training
            if self._auto_pipelines_parameters.get('data_join_only'):
                self._wml_pipeline_metadata[self._wml_client.pipelines.ConfigurationMetaNames.DOCUMENT][
                    'pipelines'][0]['nodes'] = []
                self._wml_pipeline_metadata[self._wml_client.pipelines.ConfigurationMetaNames.DOCUMENT][
                    'runtimes'] = []
            # --- end note
            else:
                # note: update autoai KB POD metadata
                self._wml_pipeline_metadata[self._wml_client.pipelines.ConfigurationMetaNames.DOCUMENT][
                    'pipelines'][0]['nodes'][0]['inputs'] = [{"id": "kb_input",
                                                              "links": [{"node_id_ref": "obm",
                                                                         "port_id_ref": "obm_out"
                                                                         }]
                                                              }]

                self._wml_pipeline_metadata[self._wml_client.pipelines.ConfigurationMetaNames.DOCUMENT][
                    'pipelines'][0]['nodes'][0]['outputs'] = [{"id": "outputData"
                                                               }]

                self._wml_pipeline_metadata[self._wml_client.pipelines.ConfigurationMetaNames.DOCUMENT][
                    'pipelines'][0]['nodes'][0]['id'] = 'kb'

                self._wml_pipeline_metadata[self._wml_client.pipelines.ConfigurationMetaNames.DOCUMENT][
                    'pipelines'][0]['nodes'][0]['runtime_ref'] = 'kb'

                self._wml_pipeline_metadata[self._wml_client.pipelines.ConfigurationMetaNames.DOCUMENT][
                    'runtimes'][0]['id'] = 'kb'
                # --- end note

            # note: insert OBM experiment metadata into the first place of nodes
            self._wml_pipeline_metadata[self._wml_client.pipelines.ConfigurationMetaNames.DOCUMENT][
                'pipelines'][0]['nodes'].insert(0, self._auto_pipelines_parameters['data_join_graph'].to_dict())
            # --- end note

            # note: insert OBM runtime metadata into the first place
            self._wml_pipeline_metadata[self._wml_client.pipelines.ConfigurationMetaNames.DOCUMENT][
                'runtimes'].insert(0, {"id": "obm",
                                       "name": "auto_ai.obm",
                                       "app_data": {
                                           "wml_data": {
                                               "hardware_spec": {}
                                           }
                                       }
                                       }
                                   )

            # note: in new v4 api we have missing hw_spec name
            t_size = self._auto_pipelines_parameters['data_join_graph'].t_shirt_size
            if len(t_size) == 1:
                self._wml_pipeline_metadata[self._wml_client.pipelines.ConfigurationMetaNames.DOCUMENT]['runtimes'][0][
                    'app_data']['wml_data']['hardware_spec']['name'] = t_size.upper()

            else:
                self._wml_pipeline_metadata[self._wml_client.pipelines.ConfigurationMetaNames.DOCUMENT]['runtimes'][0][
                    'app_data']['wml_data']['hardware_spec']['id'] = t_size
            # --- end note
            # --- end note

            if not self._wml_client.ICP:
                self._wml_pipeline_metadata[
                    self._wml_client.pipelines.ConfigurationMetaNames.DOCUMENT]['runtimes'][0]['version'] = \
                    self._auto_pipelines_parameters['obm_pod_version']
        # --- end note

        # note: KB part: send version autoai_pod_version if set in pipeline details details (See supported formats)
        if not self._wml_client.ICP and self._auto_pipelines_parameters['autoai_pod_version'] is not None:
            self._wml_pipeline_metadata[
                self._wml_client.pipelines.ConfigurationMetaNames.DOCUMENT]['runtimes'][0]['version'] = \
                self._auto_pipelines_parameters['autoai_pod_version']
        # --- end note

    def _initialize_wml_training_metadata(self,
                                          training_data_reference: List['DataConnection'],
                                          training_results_reference: 'DataConnection') -> None:
        """
        Initialization of wml training metadata (WML client Meta Parameter).

        Parameters
        ----------
        training_data_reference: List[DataConnection], required
            Data storage connection details to inform where training data is stored.

        training_results_reference: DataConnection, required
            Data storage connection details to store pipeline training results.

        """
        self._wml_training_metadata = {
            self._wml_client.training.ConfigurationMetaNames.TAGS:
                [{'value': 'autoai'}],

            self._wml_client.training.ConfigurationMetaNames.TRAINING_DATA_REFERENCES:
                [connection._to_dict() for connection in training_data_reference],

            self._wml_client.training.ConfigurationMetaNames.TRAINING_RESULTS_REFERENCE:
                training_results_reference._to_dict(),

            self._wml_client.training.ConfigurationMetaNames.PIPELINE:
                {'id': self._wml_stored_pipeline_details['metadata']['id']},

            self._wml_client.training.ConfigurationMetaNames.DESCRIPTION:
                f"{self._auto_pipelines_parameters.get('name', 'Default name.')} - "
                f"wml pipeline: {self._wml_stored_pipeline_details['metadata']['id']}",

            self._wml_client.training.ConfigurationMetaNames.NAME:
                f"{self._auto_pipelines_parameters.get('desc', 'Default description')} - "
                f"wml pipeline: {self._wml_stored_pipeline_details['metadata']['id']}",
        }

    ##########################################################
    #   WML Pipeline Part / AutoPipelineOptimizer Init Part  #
    ##########################################################
    def initiate_remote_resources(self, params: dict) -> None:
        """
        Initializes the AutoPipelines with supplied parameters.

        Parameters
        ----------
        params: dict, required
            AutoAi experiment parameters.
        """
        self._auto_pipelines_parameters = params
        self._initialize_wml_pipeline_metadata()
        self._wml_stored_pipeline_details = self._wml_client.pipelines.store(meta_props=self._wml_pipeline_metadata)

    def get_params(self) -> dict:
        """
        Get configuration parameters of AutoPipelineOptimizer.

        Returns
        -------
        Dictionary with AutoPipelineOptimizer parameters.
        """
        if self._auto_pipelines_parameters is not None:
            self._auto_pipelines_parameters['run_id'] = self._current_run_id

        return self._auto_pipelines_parameters

    ##########################################################
    #   WML Training Part / AutoPipelineOptimizer Run Part   #
    ##########################################################
    def fit(self,
            training_data_reference: List['DataConnection'],
            training_results_reference: 'DataConnection',
            background_mode: bool = True) -> dict:
        """
        Run a training process on WML of autoai on top of the training data referenced by training_data_connection.

        Parameters
        ----------
        training_data_reference: List[DataConnection], required
            Data storage connection details to inform where training data is stored.

        training_results_reference: DataConnection, required
            Data storage connection details to store pipeline training results.

        background_mode: bool, optional
            Indicator if fit() method will run in background (async) or (sync).

        Returns
        -------
        Dictionary with run details.
        """
        self._initialize_wml_training_metadata(training_data_reference=training_data_reference,
                                               training_results_reference=training_results_reference)

        # note: support for spark hummingbird
        if 'auto_ai.obm' in str(self._wml_pipeline_metadata):
            change_client_get_headers(client=self._wml_client.training._client,
                                      _type='new',
                                      workspace=self.workspace)

        run_params = self._wml_client.training.run(meta_props=self._wml_training_metadata,
                                                   asynchronous=True)

        # note: support for spark hummingbird
        if 'auto_ai.obm' in str(self._wml_pipeline_metadata):
            change_client_get_headers(client=self._wml_client.training._client,
                                      _type='old',
                                      workspace=self.workspace)

        self._current_run_id = run_params['metadata']['guid']

        wml_pipeline_details = self.get_params()

        if background_mode:
            return self._wml_client.training.get_details(training_uid=self._current_run_id)

        else:

            # TODO: Watson Studio does not support ipython widgets, this should be removed in the future
            if is_ipython() and False:
                # note: only notebook version
                total = 100 * int(wml_pipeline_details.get('max_num_daub_ensembles', 1))
                end = False
                progress_bar_1 = ProgressBar(desc="Total", total=total, position=0, ncols='100%')
                while not end:
                    progress_bar_2 = ProgressBar(desc="Waiting", total=total, leave=False, ncols='100%')
                    run_details_generator = RunDetailsGenerator(wml_client=self._wml_client,
                                                                run_id=self._current_run_id,
                                                                wait_time=0.5)

                    for message in run_details_generator:
                        progress_bar_2.set_description(desc=message)
                        if total - progress_bar_2.counter <= 2:
                            pass

                        else:
                            progress_bar_2.increment_counter(progress=1)
                            progress_bar_2.update()

                        if self._current_run_id in message:
                            end = True
                            progress_bar_1.last_update()
                            progress_bar_1.set_description(desc=message)
                            break

                    progress_bar_2.last_update()
                    progress_bar_2.close()
                    if total - progress_bar_1.counter <= 10:
                        pass

                    else:
                        progress_bar_1.increment_counter(progress=5)
                        progress_bar_1.update()
                progress_bar_1.close()

            # note: only console version
            else:
                total = 200 * int(wml_pipeline_details.get('max_num_daub_ensembles', 1))
                if self._auto_pipelines_parameters['data_join_graph']:
                    total = total * 2  # make progress bar larger when we are using OBM + KB

                progress_bar = ProgressBar(desc="Total", total=total, position=0, ncols=100)
                end = False
                while not end:
                    run_details_generator = RunDetailsGenerator(wml_client=self._wml_client,
                                                                run_id=self._current_run_id,
                                                                wait_time=0.5)

                    for message in run_details_generator:
                        progress_bar.set_description(desc=message)
                        if total - progress_bar.counter <= 5:
                            pass

                        else:
                            progress_bar.increment_counter(progress=1)
                            progress_bar.update()

                        if self._current_run_id in message:
                            end = True
                            progress_bar.last_update()
                            progress_bar.close()
                            break

            return self._wml_client.training.get_details(training_uid=self._current_run_id)

    def get_run_status(self) -> str:
        """
        Check status/state of initialized AutoPipelineOptimizer run if ran in background mode

        Returns
        -------
        Dictionary with run status details.
        """
        return self._wml_client.training.get_status(training_uid=self._current_run_id).get('state')

    def get_run_details(self) -> dict:
        """
        Get fit/run details.

        Returns
        -------
        Dictionary with AutoPipelineOptimizer fit/run details.
        """

        details = self._wml_client.training.get_details(training_uid=self._current_run_id)
        if details['entity']['status'].get('metrics', False):
            del details['entity']['status']['metrics']
            return details
        else:
            return details

    #################################################################################
    #   WML Auto_AI Trained Pipelines Part / AutoPipelineOptimizer Pipelines Part   #
    #################################################################################
    def summary(self) -> 'DataFrame':
        """
        Prints AutoPipelineOptimizer Pipelines details (autoai trained pipelines).

        Returns
        -------
        Pandas DataFrame with computed pipelines and ML metrics.
        """
        if self._current_run_id is None:
            raise FitNeeded(reason="To list computed pipelines, first schedule a fit job by using a fit() method.")

        details = self._wml_client.training.get_details(training_uid=self._current_run_id)
        return create_summary(details=details, scoring=self._auto_pipelines_parameters['scoring'])

    def get_pipeline_details(self, pipeline_name: str = None) -> dict:
        """
        Fetch specific pipeline details, eg. steps etc.

        Parameters
        ----------
        pipeline_name: str, optional
            Pipeline name eg. Pipeline_1, if not specified, best pipeline parameters will be fetched

        Returns
        -------
        Dictionary with pipeline parameters.
        """

        if self._current_run_id is None:
            raise FitNeeded(reason="To list computed pipelines parameters, "
                                   "first schedule a fit job by using a fit() method.")

        if pipeline_name is None:
            pipeline_name = self.summary().index[0]
        run_params = self._wml_client.training.get_details(training_uid=self._current_run_id)

        pipeline_parameters = {
            "composition_steps": [],
            "pipeline_nodes": [],
            "features_importance": None,
        }

        # note: retrieve all additional pipeline evaluation information
        for pipeline in run_params['entity']['status'].get('metrics', []):
            if pipeline['context']['intermediate_model']['name'].split('P')[-1] == pipeline_name.split('_')[-1]:

                pipeline_parameters['composition_steps'] = pipeline['context']['intermediate_model'][
                    'composition_steps']
                pipeline_parameters['pipeline_nodes'] = pipeline['context']['intermediate_model']['pipeline_nodes']
                pipeline_parameters['ml_metrics'] = self._get_metrics(pipeline)

                if (self._auto_pipelines_parameters['prediction_type'] == 'binary'
                        or (self._auto_pipelines_parameters['prediction_type'] == 'classification'
                            and 'binary_classification' in str(run_params))):

                    pipeline_parameters['features_importance'] = self._get_features_importance(pipeline)
                    pipeline_parameters['confusion_matrix'] = self._get_confusion_matrix(pipeline)
                    pipeline_parameters['roc_curve'] = self._get_roc_curve(pipeline)

                elif (self._auto_pipelines_parameters['prediction_type'] == 'multiclass'
                      or (self._auto_pipelines_parameters['prediction_type'] == 'classification'
                          and 'multi_class_classification' in str(run_params))):

                    pipeline_parameters['features_importance'] = self._get_features_importance(pipeline)
                    pipeline_parameters['roc_curve'] = self._get_one_vs_all(pipeline, 'roc_curve')
                    pipeline_parameters['confusion_matrix'] = self._get_one_vs_all(pipeline, 'confusion_matrix')

                elif self._auto_pipelines_parameters['prediction_type'] == 'regression':
                    pipeline_parameters['features_importance'] = self._get_features_importance(pipeline)

                else:
                    raise InvalidPredictionType(self._auto_pipelines_parameters['prediction_type'],
                                                reason="Prediction type not recognized.")

        # --- end note

        return pipeline_parameters

    @staticmethod
    def _get_metrics(pipeline_details: Dict) -> 'DataFrame':
        """Retrieve evaluation metrics data from particular pipeline details."""
        data = pipeline_details['ml_metrics']
        columns = [field for field in data.keys()]
        values = [[value for value in data.values()]]

        metrics = DataFrame(data=values, columns=columns)
        metrics.index = ['score']
        metrics = metrics.transpose()

        return metrics

    @staticmethod
    def _get_confusion_matrix(pipeline_details: Dict) -> 'DataFrame':
        """Retrieve confusion matrix data from particular pipeline details. (Binary)"""
        data = pipeline_details['context'].get('binary_classification',
                                               {'confusion_matrix': [{}, {}]})['confusion_matrix']
        columns = [name for name in data[0].keys()]
        values = [[value for value in item.values()] for item in data]

        confusion_matrix_data = DataFrame(data=values, columns=columns)
        confusion_matrix_data.index = confusion_matrix_data['true_class']
        confusion_matrix_data.drop(['true_class'], axis=1, inplace=True)

        return confusion_matrix_data

    @staticmethod
    def _get_features_importance(pipeline_details: Dict) -> 'DataFrame':
        """Retrieve features importance data from particular pipeline details."""
        data = pipeline_details['context'].get('features_importance', [{'features': {}}])[0]['features']
        columns = [name for name in data.keys()]
        values = [[value for value in data.values()]]

        features_importance_data = DataFrame(data=values, columns=columns, index=['features_importance'])
        features_importance_data = features_importance_data.transpose()
        features_importance_data.sort_values(by='features_importance', ascending=False, inplace=True)

        return features_importance_data

    @staticmethod
    def _get_roc_curve(pipeline_details: Dict) -> 'DataFrame':
        """Retrieve roc curve data from particular pipeline details. (Binary)"""
        data = pipeline_details['context'].get('binary_classification', {'roc_curve': []})['roc_curve']
        true_classes = [item['true_class'] for item in data]
        values = [data[0]['fpr'], data[1]['fpr'], data[0]['thresholds'], data[1]['thresholds'], data[0]['tpr'],
                  data[1]['tpr']]

        index = MultiIndex.from_product([['fpr', 'thresholds', 'tpr'], true_classes],
                                        names=['measurement types', 'true_class'])

        roc_curve = DataFrame(data=values, index=index)

        return roc_curve

    @staticmethod
    def _get_one_vs_all(pipeline_details: Dict, option: str) -> 'DataFrame':
        """Retrieve roc curve or confusion matrix data from particular pipeline details. (Multiclass)"""
        data = pipeline_details['context'].get('multi_class_classification', {'one_vs_all': []})['one_vs_all']

        if option == 'confusion_matrix':
            confusion_matrix_values = [item['confusion_matrix'].values() for item in data]
            confusion_matrix_data = DataFrame(data=confusion_matrix_values,
                                              columns=['fn', 'fp', 'tn', 'tp', 'true_class'])
            confusion_matrix_data.index = confusion_matrix_data['true_class']
            confusion_matrix_data.drop(['true_class'], axis=1, inplace=True)

            return confusion_matrix_data

        elif option == 'roc_curve':
            true_classes = [item['roc_curve']['true_class'] for item in data]
            values = [item['roc_curve']['fpr'] for item in data] + [item['roc_curve']['thresholds'] for item in
                                                                    data] + [item['roc_curve']['tpr'] for item in data]
            index = MultiIndex.from_product([['fpr', 'thresholds', 'tpr'], true_classes],
                                            names=['measurement types', 'true_class'])
            roc_curve = DataFrame(data=values, index=index)

            return roc_curve

        else:
            raise ValueError(f"Wrong option for \"one_vs_all\": {option}. "
                             f"Available options: [confusion_matrix, roc_curve]")

    def get_pipeline(self, pipeline_name: str, local_path: str = '.', persist: 'bool' = False) -> 'Pipeline':
        """
        Download specified pipeline from WML.

        Parameters
        ----------
        pipeline_name: str, required
            Pipeline name, if you want to see the pipelines names, please use summary() method.

        local_path: str, optional
            Local filesystem path, if not specified, current directory is used.

        persist: bool, optional
            Indicates if selected pipeline should be stored locally.

        Returns
        -------
        Scikit-Learn pipeline.
        """
        try_import_autoai_libs()
        check_dependencies_versions(lib_name='scikit', error=True)
        check_dependencies_versions(lib_name='numpy', error=True)

        run_params = self._wml_client.training.get_details(training_uid=self._current_run_id)
        try:
            pipelines: Dict[str, 'Pipeline'] = fetch_pipelines(
                run_params=run_params,
                path=local_path,
                pipeline_name=pipeline_name,
                load_pipelines=True,
                store=persist,
                wml_client=self._wml_client if self._wml_client.ICP else None
            )

        except Exception as e:
            raise PipelineNotLoaded(
                pipeline_name,
                reason=f'Fit finished but there was some error during loading a pipeline from WML. Error: {e}')

        return pipelines.get(pipeline_name)

    def get_best_pipeline(self, local_path: str = '.', persist: 'bool' = False) -> 'Pipeline':
        """
        Download best pipeline from WML.

        Parameters
        ----------
        local_path: str, optional
            Local filesystem path, if not specified, current directory is used.

        persist: bool, optional
            Indicates if selected pipeline should be stored locally.

        Returns
        -------
        Scikit-Learn pipeline.
        """
        try_import_autoai_libs()
        check_dependencies_versions(lib_name='scikit', error=True)
        check_dependencies_versions(lib_name='numpy', error=True)

        best_pipeline_name = self.summary().index[0]
        try:
            pipelines: Dict[str, 'Pipeline'] = fetch_pipelines(
                run_params=self._wml_client.training.get_details(training_uid=self._current_run_id),
                path=local_path,
                pipeline_name=best_pipeline_name,
                load_pipelines=True,
                store=persist,
                wml_client=self._wml_client if self._wml_client.ICP else None
            )

        except Exception as e:
            raise PipelineNotLoaded(
                'Best Pipeline',
                reason=f'Fit finished but there was some error during loading a pipeline from WML. Error: {e}')

        return pipelines.get(best_pipeline_name)



################################################################################
#
# Licensed Materials - Property of IBM
# (C) Copyright IBM Corp. 2017
# US Government Users Restricted Rights - Use, duplication disclosure restricted
# by GSA ADP Schedule Contract with IBM Corp.
#
################################################################################

from  ibm_watson_machine_learning.libs.repo.mlrepositoryartifact.scikit_artifact_loader import ScikitArtifactLoader


class ScikitPipelineModelLoader(ScikitArtifactLoader):
    """
        Returns pipeline model instance associated with this model artifact.

        :return: pipeline model
        :rtype: scikit.learn.Pipeline
        """
    def model_instance(self,as_type=None):
        """
           :param as_type: string type referring to the model type to be returned.
           This parameter is applicable for xgboost models only.
           Currently accepts:
             'Booster': returns a model of type xgboost.Booster
             'XGBRegressor': returns a model of type xgboost.sklearn.XGBRegressor
           :return: returns a scikit model or an xgboost model of type xgboost.Booster
           or xgboost.sklearn.XGBRegressor
         """
        return self.load(as_type)



import unittest
from preparation_and_cleaning import *
from models_preparation import *
from ibm_watson_machine_learning.utils.log_util import get_logger


class TestRuntimeSpec(unittest.TestCase):
    runtime_uid = None
    logger = get_logger(__name__)

    @classmethod
    def setUpClass(self):
        TestRuntimeSpec.logger.info("Service Instance: setting up credentials")
        self.wml_credentials = get_wml_credentials()
        self.client = get_client()
        self.custom_library_path = os.path.join(os.getcwd(), 'artifacts', 'scikit_xgboost_model.tar.gz') # TODO

    def test_1_service_instance_details(self):
        TestRuntimeSpec.logger.info("Check client ...")
        self.assertTrue(self.client.__class__.__name__ == 'APIClient')

        TestRuntimeSpec.logger.info("Getting instance details ...")
        details = self.client.service_instance.get_details()

        TestRuntimeSpec.logger.debug(details)
        self.assertTrue("published_models" in str(details))
        self.assertEqual(type(details), dict)

    def test_2_create_runtime(self):
        meta = {
            self.client.runtimes.ConfigurationMetaNames.NAME: "runtime_spec_python_3.5_new",
            self.client.runtimes.ConfigurationMetaNames.PLATFORM: {
                "name": "python",
                "version": "3.6"
            }
        }
        runtime_details = self.client.runtimes.store(meta)
        TestRuntimeSpec.runtime_uid = self.client.runtimes.get_uid(runtime_details)
        runtime_url = self.client.runtimes.get_href(runtime_details)

        self.assertTrue(TestRuntimeSpec.runtime_uid is not None)

    def test_3_get_details(self):
        self.client.runtimes.get_details(TestRuntimeSpec.runtime_uid)

    def test_4_list(self):
        self.client.runtimes.list()

    def test_5_delete_runtime(self):
        self.client.runtimes.delete(TestRuntimeSpec.runtime_uid)


if __name__ == '__main__':
    unittest.main()



import unittest

from ibm_watson_machine_learning.utils.log_util import get_logger
from preparation_and_cleaning import *
from models_preparation import *


class TestAIFunction(unittest.TestCase):
    runtime_uid = None
    lib_uid = None
    deployment_uid = None
    pipeline_uid = None
    scoring_url = None
    logger = get_logger(__name__)

    @classmethod
    def setUpClass(self):
        TestAIFunction.logger.info("Service Instance: setting up credentials")
        self.lib_filepath = os.path.join(os.getcwd(), 'artifacts', 'ai_function.gz')
        self.wml_credentials = get_wml_credentials()
        self.client = get_client()

        self.function_name = 'sample v4 pipeline'
        self.deployment_name = "Test deployment"

    def test_01_service_instance_details(self):
        TestAIFunction.logger.info("Check client ...")
        self.assertTrue(self.client.__class__.__name__ == 'APIClient')

        TestAIFunction.logger.info("Getting instance details ...")
        details = self.client.service_instance.get_details()
        TestAIFunction.logger.debug(details)

        self.assertTrue("published_models" in str(details))
        self.assertEqual(type(details), dict)

    def test_02_create_pipeline(self):

        self.client.repository.PipelineMetaNames.show()
        lib_meta = {
            self.client.runtimes.LibraryMetaNames.NAME: "sample v4 pipeline2",
            self.client.runtimes.LibraryMetaNames.PLATFORM:  {"name": "python","versions": ["3.6"]},
            self.client.runtimes.LibraryMetaNames.VERSION: "1",
            self.client.runtimes.LibraryMetaNames.FILEPATH: TestAIFunction.lib_filepath
        }
        lib_details = self.client.runtimes.store_library(lib_meta)
        TestAIFunction.lib_uid = self.client.runtimes.get_library_uid(lib_details)
        metadata = {
            self.client.pipelines.ConfigurationMetaNames.NAME: "sample v4 pipeline",
            self.client.pipelines.ConfigurationMetaNames.DESCRIPTION: "sample description",
            self.client.pipelines.ConfigurationMetaNames.DOCUMENT: {
                "doc_type": "pipeline",
                "version": "2.0",
                "primary_pipeline": "dlaas_only",
                "pipelines": [
                    {
                        "id": "dlaas_only",
                        "runtime_ref": "hybrid",
                        "nodes": [
                            {
                                "id": "training",
                                "type": "model_node",
                                "op": "dl_train",
                                "runtime_ref": "DL",
                                "inputs": [
                                ],
                                "outputs": [],
                                "parameters": {
                                    "name": "tf-mnist",
                                    "description": "Simple MNIST model implemented in TF",
                                    "command": "python3 convolutional_network.py --trainImagesFile ${DATA_DIR}/train-images-idx3-ubyte.gz --trainLabelsFile ${DATA_DIR}/train-labels-idx1-ubyte.gz --testImagesFile ${DATA_DIR}/t10k-images-idx3-ubyte.gz --testLabelsFile ${DATA_DIR}/t10k-labels-idx1-ubyte.gz --learningRate 0.001 --trainingIters 6000",
                                    "model_definition_url": "/v4/libraries"+TestAIFunction.lib_uid,
                                    "compute_configuration_name": "k80",
                                    "compute_configuration_nodes": 1,
                                    "target_bucket": "wml-dev-results"
                                }
                            }
                        ]
                    }
                ],
                "schemas": [
                    {
                        "id": "schema1",
                        "fields": [
                            {
                                "name": "text",
                                "type": "string"
                            }
                        ]
                    }
                ],
                "runtimes": [
                    {
                        "id": "DL",
                        "name": "tensorflow",
                        "version": "1.5-py3"
                    }
                ]
            }

        }
        pipeline_details = self.client.pipelines.store(metadata)
        TestAIFunction.pipeline_uid = self.client.pipelines.get_uid(pipeline_details)



    def test_03_update_pipeline(self):
        pipeline_props = {
            self.client.repository.FunctionMetaNames.NAME: 'sample_v4_pipeline_new',
        }

        details = self.client.pipelines.update(TestAIFunction.pipeline_uid, pipeline_props)
        self.assertFalse('sample v4 pipeline' in json.dumps(details))

    def test_05_get_details(self):

        details = self.client.pipelines.get_details(self.pipeline_uid)
        self.assertTrue('sample_v4_pipeline_new' in str(details))

    def test_06_list(self):
        self.client.pipelines.list()

    def test_12_delete_pipeline(self):

        TestAIFunction.logger.info("Delete function")
        self.client.pipelines.delete(TestAIFunction.pipeline_uid)
        self.client.runtimes.delete_library(TestAIFunction.lib_uid)


if __name__ == '__main__':
    unittest.main()



__all__ = [
    "AutoPipelinesRuns"
]

from copy import deepcopy
from typing import List, Dict, Union, Optional

from pandas import DataFrame

from ibm_watson_machine_learning.experiment.autoai.engines import WMLEngine
from ibm_watson_machine_learning.experiment.autoai.optimizers import RemoteAutoPipelines
from ibm_watson_machine_learning.preprocessing import DataJoinGraph
from ibm_watson_machine_learning.utils.autoai.utils import NextRunDetailsGenerator, get_node_and_runtime_index
from ibm_watson_machine_learning.helpers import DataConnection
from .base_auto_pipelines_runs import BaseAutoPipelinesRuns
from ibm_watson_machine_learning.wml_client_error import ApiRequestFailure


class AutoPipelinesRuns(BaseAutoPipelinesRuns):
    """
    AutoPipelinesRuns class is used to work with historical Optimizer runs.

    Parameters
    ----------
    engine: WMLEngine, required
        WMLEngine to handle WML operations.

    filter: str, optional
        Filter, user can choose which runs to fetch specifying AutoPipelines name.
    """

    def __init__(self, engine: 'WMLEngine', filter: str = None) -> None:
        self._wml_engine: 'WMLEngine' = engine
        self.auto_pipeline_optimizer_name = filter
        self._workspace = None

    def __call__(self, *, filter: str) -> 'AutoPipelinesRuns':
        self.auto_pipeline_optimizer_name = filter
        return self

    def list(self) -> 'DataFrame':
        """
        Lists historical runs/fits with status. If user has a lot of runs stored in the WML,
        it may take long time to fetch all the information.

        Returns
        -------
        Pandas DataFrame with runs IDs and state.
        """

        columns = ['timestamp', 'run_id', 'state', 'auto_pipeline_optimizer name']
        wml_pipelines_names = []
        wml_pipelines_types = []

        # note: download all runs details
        runs_details = self._wml_engine._wml_client.training.get_details(limit=50)
        data = runs_details.get('resources', [])
        run_details_generator = NextRunDetailsGenerator(wml_client=self._wml_engine._wml_client,
                                                        href=runs_details.get('next', {'href': None}).get('href'))
        for entry in run_details_generator:
            data.extend(entry)
        # --- end note

        # note: some of the pending experiments do not have these information (checking with if statement)
        runs_pipeline_ids = [run['entity']['pipeline']['id'] for run in data if
                             run['entity'].get('pipeline', {}).get('id')]
        runs_timestamps = [run['metadata'].get('modified_at') for run in data if
                           run['entity'].get('pipeline', {}).get('id')]
        data = [run for run in data if run['entity'].get('pipeline', {}).get('id')]
        # --- end note

        for wml_pipeline_id in runs_pipeline_ids:
            pipeline_details = self._wml_engine._wml_client.pipelines.get_details(
                pipeline_uid=wml_pipeline_id)

            wml_pipeline_type = ('autoai' if 'automl' in str(pipeline_details) and
                                             'hybrid' in str(pipeline_details) else 'other')

            try:
                wml_pipeline_name = pipeline_details['metadata'].get('name')

            except ApiRequestFailure as e:
                wml_pipeline_name = 'Unknown'

            wml_pipelines_names.append(wml_pipeline_name)
            wml_pipelines_types.append(wml_pipeline_type)

        if self.auto_pipeline_optimizer_name is not None:
            values = [[timestamp, run['metadata']['guid'], run['entity']['status']['state'], wml_pipeline_name] for
                      timestamp, run, wml_pipeline_name, wml_pipeline_type in zip(runs_timestamps,
                                                                                  data,
                                                                                  wml_pipelines_names,
                                                                                  wml_pipelines_types)
                      if wml_pipeline_name == self.auto_pipeline_optimizer_name and wml_pipeline_type == 'autoai']

        else:
            values = [[timestamp, run['metadata']['guid'], run['entity']['status']['state'], wml_pipeline_name] for
                      timestamp, run, wml_pipeline_name, wml_pipeline_type in zip(runs_timestamps,
                                                                                  data,
                                                                                  wml_pipelines_names,
                                                                                  wml_pipelines_types)
                      if wml_pipeline_type == 'autoai']

        runs = DataFrame(data=values, columns=columns)
        return runs.sort_values(by=["timestamp"], ascending=False)

    def get_params(self, run_id: str = None) -> dict:
        """
        Get executed optimizers configs parameters based on the run_id.

        Parameters
        ----------
        run_id: str, optional
            ID of the fit/run. If not specified, latest is taken.

        Returns
        -------
        Dictionary with optimizer configuration parameters.

        Example
        -------
        >>> from ibm_watson_machine_learning.experiment import AutoAI
        >>> experiment = AutoAI(credentials, ...)
        >>>
        >>> experiment.runs.get_params(run_id='02bab973-ae83-4283-9d73-87b9fd462d35')
        >>> experiment.runs.get_params()
            {
                'name': 'test name',
                'desc': 'test description',
                'prediction_type': 'classification',
                'prediction_column': 'y',
                'scoring': 'roc_auc',
                'test_size': 0.1,
                'max_num_daub_ensembles': 1
            }
        """

        if run_id is None:
            optimizer_id = self._wml_engine._wml_client.training.get_details(
                limit=1
            ).get('resources')[0]['entity']['pipeline']['id']

        else:
            optimizer_id = self._wml_engine._wml_client.training.get_details(
                training_uid=run_id
            ).get('entity')['pipeline']['id']

        optimizer_config = self._wml_engine._wml_client.pipelines.get_details(pipeline_uid=optimizer_id)

        # note: if experiment has more than 1 node (e.g. OBM + KB), we need to find which one is KB
        kb_node_number, kb_runtime_number = get_node_and_runtime_index(node_name='kb',
                                                                       optimizer_config=optimizer_config)
        # --- end note

        # note: try to find obm node
        obm_node_number, obm_runtime_number = get_node_and_runtime_index(node_name='obm',
                                                                         optimizer_config=optimizer_config)
        # --- end note

        try:
            name = optimizer_config['entity']['name']
            description = optimizer_config['entity'].get('description', '')

        except KeyError:
            name = optimizer_config['metadata']['name']
            description = optimizer_config['metadata'].get('description', '')

        # note: check if not only data preprocessing experiment
        if kb_node_number is not None:
            kb_parameters = optimizer_config['entity']['document']['pipelines'][0]['nodes'][kb_node_number][
                'parameters']['optimization']
            kb_wml_data = optimizer_config['entity']['document']['runtimes'][kb_runtime_number]['app_data']['wml_data']

            csv_separator = optimizer_config['entity']['document']['pipelines'][0]['nodes'][kb_node_number][
                'parameters'].get('input_file_separator', ',')
            excel_sheet = optimizer_config['entity']['document']['pipelines'][0]['nodes'][kb_node_number][
                'parameters'].get('excel_sheet', 0)

            params = {
                'name': name,
                'desc': description,
                'prediction_type': kb_parameters['learning_type'],
                'prediction_column': kb_parameters['label'],
                'scoring': kb_parameters['scorer_for_ranking'],
                'test_size': kb_parameters['holdout_param'],
                'max_num_daub_ensembles': kb_parameters['max_num_daub_ensembles'],
                't_shirt_size': kb_wml_data['hardware_spec']['id'],
                'daub_include_only_estimators': kb_parameters.get('daub_include_only_estimators'),
                'train_sample_rows_test_size': kb_parameters.get('train_sample_rows_test_size'),
                'csv_separator': csv_separator,
                'excel_sheet': excel_sheet
            }

            if kb_parameters.get('train_sample_rows_test_size'):
                params['train_sample_rows_test_size'] = kb_parameters['train_sample_rows_test_size']
        else:
            params = {
                'name': name,
                'desc': description,
                'prediction_type': None,
                'prediction_column': None,
                'scoring': None,
                'data_join_only': True
            }
        # --- end note

        # note: rebuild DataJoinGraph object from historical run
        if obm_node_number is not None:
            obm_parameters = optimizer_config['entity']['document']['pipelines'][0]['nodes'][obm_node_number]
            data_join_graph = DataJoinGraph._from_dict(_dict=obm_parameters)

            obm_wml_data = optimizer_config['entity']['document']['runtimes'][obm_runtime_number]['app_data'][
                'wml_data']
            data_join_graph.t_shirt_size = obm_wml_data['hardware_spec']['name']

            params['data_join_graph'] = data_join_graph
        # --- end note

        return params

    def get_run_details(self, run_id: str = None) -> dict:
        """
        Get run details. If run_id is not supplied, last run will be taken.

        Parameters
        ----------
        run_id: str, optional
            ID of the fit/run.

        Returns
        -------
        Dictionary with run configuration parameters.

        Example
        -------
        >>> from ibm_watson_machine_learning.experiment import AutoAI
        >>> experiment = AutoAI(credentials, ...)
        >>>
        >>> experiment.runs.get_run_details(run_id='02bab973-ae83-4283-9d73-87b9fd462d35')
        >>> experiment.runs.get_run_details()
        """
        if run_id is None:
            details = self._wml_engine._wml_client.training.get_details(limit=1).get('resources')[0]

        else:
            details = self._wml_engine._wml_client.training.get_details(training_uid=run_id)

        if details['entity']['status'].get('metrics', False):
            del details['entity']['status']['metrics']
            return details
        else:
            return details

    def get_optimizer(self,
                      run_id: Optional[str] = None,
                      metadata: Dict[str, Union[List['DataConnection'],  'DataConnection', str, int]] = None
                      ) -> 'RemoteAutoPipelines':
        """
        Creates instance of AutoPipelinesRuns with all computed pipelines computed by AutoAi on WML.

        Parameters
        ----------
        run_id: str, optional
            ID of the fit/run.

        metadata: dict, optional
            Option to pass information about COS data reference or WSD (auto-gen notebook)

        Returns
        -------
        AutoPipelinesRuns class instance.

        Example
        -------
        >>> from ibm_watson_machine_learning.experiment import AutoAI
        >>> experiment = AutoAI(credentials, ...)
        >>>
        >>> historical_optimizer = experiment.runs.get_optimizer(run_id='02bab973-ae83-4283-9d73-87b9fd462d35')
        """
        # note: normal scenario
        if metadata is None:
            optimizer_parameters = self.get_params(run_id=run_id)

            remote_pipeline_optimizer = RemoteAutoPipelines(**optimizer_parameters, engine=self._wml_engine)

            remote_pipeline_optimizer._engine._current_run_id = run_id
            remote_pipeline_optimizer._workspace = self._workspace

            return remote_pipeline_optimizer
        # --- end note

        # note: WSD auto-gen notebook scenario (when user provides his WMLS credentials)
        else:
            from ibm_watson_machine_learning.experiment import AutoAI
            training_result_reference = metadata.get('training_result_reference')
            run_id = training_result_reference.location.path.split('/')[-3]

            return AutoAI(self._workspace.wml_credentials,
                          space_id=self._workspace.space_id
                          ).runs.get_optimizer(run_id)
        # --- end note

    def get_data_connections(self, run_id: str) -> List['DataConnection']:
        """
        Create DataConnection objects for further user usage
            (eg. to handle data storage connection or to recreate autoai holdout split).

        Parameters
        ----------
        run_id: str, required
            ID of the historical fit/run.

        Returns
        -------
        List['DataConnection'] with populated optimizer parameters

        Example
        -------
        >>> from ibm_watson_machine_learning.experiment import AutoAI
        >>> experiment = AutoAI(credentials, ...)
        >>>
        >>> data_connections = experiment.runs.get_data_connections(run_id='02bab973-ae83-4283-9d73-87b9fd462d35')
        """
        optimizer_parameters = self.get_params(run_id=run_id)
        training_data_references = self.get_run_details(run_id=run_id)['entity']['training_data_references']

        data_connections = [
            DataConnection._from_dict(_dict=data_connection) for data_connection in training_data_references]

        for data_connection in data_connections:  # note: populate DataConnections with optimizer params
            data_connection.auto_pipeline_params = deepcopy(optimizer_parameters)
            data_connection._wml_client = self._wml_engine._wml_client
            data_connection._run_id = run_id

        return data_connections



from ibm_watson_machine_learning.libs.repo.mlrepository import ModelArtifact


class XGBoostModelArtifact(ModelArtifact):
    """
    Class representing xgboost model artifact
    """
    def __init__(self, uid, name, meta_props):
        """
        Constructor for xgboost model artifact
        :param uid: unique id for xgboost model artifact
        :param name: name of the model
        :param metaprops: properties of the model and model artifact
        """
        super(XGBoostModelArtifact, self).__init__(uid, name, meta_props)


################################################################################
#
# Licensed Materials - Property of IBM
# (C) Copyright IBM Corp. 2017
# US Government Users Restricted Rights - Use, duplication disclosure restricted
# by GSA ADP Schedule Contract with IBM Corp.
#
################################################################################

from ibm_watson_machine_learning.libs.repo.util.unique_id_gen import uid_generate

import os, shutil


class RuntimesArtifactLoader(object):

    def download_runtimes(self, file_path):
        return self.__extract_content(file_path)

    def __extract_content(self, file_path):

        if file_path is None:
            file_path = 'runtimes_artifact'
            try:
                shutil.rmtree(file_path)
            except:
                pass
            os.makedirs(file_path)

        try:
            id_length = 20
            runtimes_generated_id = uid_generate(id_length)
            if self.runtimespec_path is not None:
                file_name = '{}/{}_{}.yml'.format(file_path, os.path.basename(self.runtimespec_path),runtimes_generated_id)
            else:
                file_name = '{}/{}_{}.yml'.format(file_path, self.name, runtimes_generated_id)
            input_stream = self._content_reader()
            file_content = input_stream.read()
            gz_f = open(file_name, 'wb+')
            gz_f.write(file_content)
            gz_f.close()
            return os.path.abspath(file_name)
        except Exception as ex:
            raise ex

    def _content_reader(self):
        if self._content_href is not None:
            if self._content_href.__contains__("runtimes"):
                return self.client.repository_api.download_artifact_content(self._content_href, 'false', accept='text/plain')



from ibm_watson_machine_learning.libs.repo.mlrepository import ModelArtifact


class HybridModelArtifact(ModelArtifact):
    """
    Class representing Hybrid model artifact
    """
    def __init__(self, uid, name, meta_props):
        """
        Constructor for Hybrid model artifact
        :param uid: unique id for Hybrid model artifact
        :param name: name of the model
        :param metaprops: properties of the model and model artifact
        """
        super(HybridModelArtifact, self).__init__(uid, name, meta_props)




import io
import time
from typing import TYPE_CHECKING, Any, Dict, Union, List, Optional
from warnings import warn

from pandas import DataFrame, concat

from .base_deployment import BaseDeployment
from ..helpers import DataConnection
from ..utils import StatusLogger, print_text_header_h1
from ..utils.autoai.utils import init_cos_client, try_load_dataset
from ..utils.autoai.watson_studio import change_client_get_headers
from ..utils.deployment.errors import BatchJobFailed, MissingScoringResults
from ..wml_client_error import WMLClientError
from ..workspace import WorkSpace
from ..utils.autoai.connection import validate_source_data_connections, validate_deployment_output_connection


if TYPE_CHECKING:
    from sklearn.pipeline import Pipeline
    from pandas import DataFrame
    from numpy import ndarray

__all__ = [
    "Batch"
]


class Batch(BaseDeployment):
    """
    The Batch Deployment class.
    With this class object you can manage any batch deployment.

    Parameters
    ----------
    source_wml_credentials: dictionary, required
        Credentials to Watson Machine Learning instance where training was performed.

    source_project_id: str, optional
        ID of the Watson Studio project where training was performed.

    source_space_id: str, optional
        ID of the Watson Studio Space where training was performed.

    target_wml_credentials: dictionary, required
        Credentials to Watson Machine Learning instance where you want to deploy.

    target_project_id: str, optional
        ID of the Watson Studio project where you want to deploy.

    target_space_id: str, optional
        ID of the Watson Studio Space where you want to deploy.
    """

    def __init__(self,
                 source_wml_credentials: Union[dict, 'WorkSpace'] = None,
                 source_project_id: str = None,
                 source_space_id: str = None,
                 target_wml_credentials: Union[dict, 'WorkSpace'] = None,
                 target_project_id: str = None,
                 target_space_id: str = None,
                 wml_credentials: Union[dict, 'WorkSpace'] = None,
                 project_id: str = None,
                 space_id: str = None):

        # note: backward compatibility
        if wml_credentials is not None:
            source_wml_credentials = wml_credentials
            warn("\"wml_credentials\" parameter is depreciated, please use \"source_wml_credentials\"")
            print("\"wml_credentials\" parameter is depreciated, please use \"source_wml_credentials\"")

        if project_id is not None:
            source_project_id = project_id
            warn("\"project_id\" parameter is depreciated, please use \"source_project_id\"")
            print("\"project_id\" parameter is depreciated, please use \"source_project_id\"")

        if space_id is not None:
            source_space_id = space_id
            warn("\"space_id\" parameter is depreciated, please use \"source_space_id\"")
            print("\"space_id\" parameter is depreciated, please use \"source_space_id\"")
        # --- end note

        # note: as workspace is not clear enough to understand, there is a possibility to use pure
        # wml credentials with project and space IDs, but in addition we
        # leave a possibility to use a previous WorkSpace implementation, it could be passed as a first argument
        super().__init__(deployment_type='batch')
        if isinstance(source_wml_credentials, WorkSpace):
            self._source_workspace = source_wml_credentials

        elif isinstance(source_wml_credentials, dict):
            self._source_workspace = WorkSpace(wml_credentials=source_wml_credentials.copy(),
                                               project_id=source_project_id,
                                               space_id=source_space_id)
        else:
            self._source_workspace = None

        if target_wml_credentials is None:
            if isinstance(source_wml_credentials, WorkSpace):
                self._target_workspace = source_wml_credentials

            elif isinstance(source_wml_credentials, dict):
                self._target_workspace = WorkSpace(wml_credentials=source_wml_credentials.copy(),
                                                   project_id=source_project_id,
                                                   space_id=source_space_id)
            else:
                self._target_workspace = None

        else:
            if isinstance(target_wml_credentials, WorkSpace):
                self._target_workspace = target_wml_credentials

            elif isinstance(target_wml_credentials, dict):
                self._target_workspace = WorkSpace(wml_credentials=target_wml_credentials.copy(),
                                                   project_id=target_project_id,
                                                   space_id=target_space_id)

                # note: only if user provides target WML information
                if self._source_workspace is None:
                    self._source_workspace = copy(self._target_workspace)

            else:
                self._target_workspace = None
        # --- end note

        self.name = None
        self.id = None
        self.asset_id = None

    def __repr__(self):
        return f"name: {self.name}, id: {self.id}, asset_id: {self.asset_id}"

    def __str__(self):
        return f"name: {self.name}, id: {self.id}, asset_id: {self.asset_id}"

    def score(self, **kwargs):
        raise NotImplementedError("Batch deployment supports only job runs.")

    def create(self,
               model: Any,
               deployment_name: str,
               metadata: Optional[Dict] = None,
               training_data: Optional[Union['DataFrame', 'ndarray']] = None,
               training_target: Optional[Union['DataFrame', 'ndarray']] = None,
               experiment_run_id: Optional[str] = None) -> None:
        """
        Create deployment from a model.

        Parameters
        ----------
        model: Union[Any, str], required
            Model object to deploy or local path to the model.

        deployment_name: str, required
            Name of the deployment

        training_data: Union['pandas.DataFrame', 'numpy.ndarray'], optional
            Training data for the model

        training_target: Union['pandas.DataFrame', 'numpy.ndarray'], optional
            Target/label data for the model

        metadata: dictionary, optional
            Model meta properties.

        experiment_run_id: str, optional
            ID of a training/experiment (only applicable for AutoAI deployments)

        Example
        -------
        >>> from ibm_watson_machine_learning.deployment import Batch
        >>>
        >>> deployment = Batch(
        >>>        wml_credentials={
        >>>              "apikey": "...",
        >>>              "iam_apikey_description": "...",
        >>>              "iam_apikey_name": "...",
        >>>              "iam_role_crn": "...",
        >>>              "iam_serviceid_crn": "...",
        >>>              "instance_id": "...",
        >>>              "url": "https://us-south.ml.cloud.ibm.com"
        >>>            },
        >>>         project_id="...",
        >>>         space_id="...")
        >>>
        >>> deployment.create(
        >>>        experiment_run_id="...",
        >>>        model=model,
        >>>        deployment_name='My new deployment'
        >>>    )
        """
        return super().create(model=model,
                              deployment_name=deployment_name,
                              metadata=metadata,
                              training_data=training_data,
                              training_target=training_target,
                              experiment_run_id=experiment_run_id,
                              deployment_type='batch')

    @BaseDeployment._project_to_space_to_project
    def get_params(self) -> Dict:
        """Get deployment parameters."""
        return super().get_params()

    @BaseDeployment._project_to_space_to_project
    def run_job(self,
                payload: Union[DataFrame, List[DataConnection]],
                output_data_reference: 'DataConnection' = None,
                background_mode: 'bool' = True) -> Union[Dict, Dict[str, List], DataConnection]:
        """
        Batch scoring job on WML. Payload or Payload data reference is required. It is passed to the WML where model have been deployed.

        Parameters
        ----------
        payload: pandas.DataFrame or List[DataConnection], required
            DataFrame with data to test the model or data storage connection details to inform where payload data is stored.

        output_data_reference: DataConnection, optional
             DataConnection to the output COS for storing predictions.
             Required nly when DataConnections are as a payload.

        background_mode: bool, optional
            Indicator if score() method will run in background (async) or (sync).

        Returns
        -------
        Dictionary with scoring job details.

        Example
        -------
        >>> score_details = deployment.score(payload=test_data)
        >>> print(score_details['entity']['scoring'])
            {'input_data': [{'fields': ['sepal_length',
                          'sepal_width',
                          'petal_length',
                          'petal_width'],
                         'values': [[4.9, 3.0, 1.4, 0.2]]}],
           'predictions': [{'fields': ['prediction', 'probability'],
                         'values': [['setosa',
                           [0.9999320742502246,
                            5.1519823540224506e-05,
                            1.6405926235405522e-05]]]}]
        >>>
        >>> payload_reference = DataConnection(location=DSLocation(asset_id=asset_id))
        >>> score_details = deployment.score(payload=payload_reference, output_data_filename = "scoring_output.csv")
        """
        # note: support for DataFrame payload
        if isinstance(payload, DataFrame):
            scoring_payload = {
                self._target_workspace.wml_client.deployments.ScoringMetaNames.INPUT_DATA: [{'values': payload}]
            }
        # note: support for DataConnections and dictionaries payload
        elif isinstance(payload, list):
            if isinstance(payload[0], DataConnection):
                payload = validate_source_data_connections(source_data_connections=payload,
                                                           workspace=self._source_workspace,
                                                           deployment=True)
                payload = [data_connection._to_dict() for data_connection in payload]
            elif isinstance(payload[0], dict):
                pass
            else:
                raise ValueError(f"Current payload type: list of {type(payload[0])} is not supported.")

            if output_data_reference is None:
                raise ValueError("\"output_data_reference\" should be provided.")

            if isinstance(output_data_reference, DataConnection):
                output_data_reference = validate_deployment_output_connection(results_data_connection=output_data_reference,
                                                                              workspace=self._target_workspace,
                                                                              source_data_connections = payload)
                output_data_reference = output_data_reference._to_dict()

            scoring_payload = {
                self._target_workspace.wml_client.deployments.ScoringMetaNames.INPUT_DATA_REFERENCES: payload,
                self._target_workspace.wml_client.deployments.ScoringMetaNames.OUTPUT_DATA_REFERENCE:
                    output_data_reference}

        else:
            raise ValueError(
                f"Incorrect payload type. Required: DataFrame or List[DataConnection], Passed: {type(payload)}")

        scoring_payload['hybrid_pipeline_hardware_specs'] = [
            {
                'node_runtime_id': 'auto_ai.kb',
                'hardware_spec': {
                    'name': 'M'
                }
            }
        ]

        if self._obm:
            scoring_payload['hybrid_pipeline_hardware_specs'].append(
                {
                    "node_runtime_id": "auto_ai.obm",
                    "hardware_spec": {
                        "name": "M",
                        "num_nodes": 2
                    }
                }
            )

        # note: support for spark hummingbird
        if self._obm:
            change_client_get_headers(client=self._target_workspace.wml_client.training._client,
                                      _type='new',
                                      workspace=self._target_workspace)

        job_details = self._target_workspace.wml_client.deployments.create_job(self.id, scoring_payload)

        # note: support for spark hummingbird
        if self._obm:
            change_client_get_headers(client=self._target_workspace.wml_client.training._client,
                                      _type='old',
                                      workspace=self._target_workspace)

        if background_mode:
            return job_details

        else:
            # note: monitor scoring job
            job_id = self._target_workspace.wml_client.deployments.get_job_uid(job_details)
            print_text_header_h1(u'Synchronous scoring for id: \'{}\' started'.format(job_id))

            status = self.get_job_status(job_id)['state']

            with StatusLogger(status) as status_logger:
                while status not in ['failed', 'error', 'completed', 'canceled']:
                    time.sleep(10)
                    status = self.get_job_status(job_id)['state']
                    status_logger.log_state(status)
            # --- end note

            if u'completed' in status:
                print(u'\nScoring job  \'{}\' finished successfully.'.format(job_id))
            else:
                raise BatchJobFailed(job_id, f"Scoring job failed with status: {self.get_job_status(job_id)}")

            return self.get_job_params(job_id)

    @BaseDeployment._project_to_space_to_project
    def rerun_job(self,
                  scoring_job_id: str,
                  background_mode: bool = True) -> Union[dict, 'DataFrame', 'DataConnection']:
        """
        Rerun scoring job with the same parameters as job described by 'scoring_job_id'.

        Parameters
        ----------
        scoring_job_id: str
           Id described scoring job.

        background_mode: bool, optional
           Indicator if score_rerun() method will run in background (async) or (sync).

        Returns
        -------
        Dictionary with scoring job details.

        Example
        -------
        >>> scoring_details = deployment.score_rerun(scoring_job_id)
        """
        scoring_params = self.get_job_params(scoring_job_id)['entity']['scoring']
        input_data_references = self._target_workspace.wml_client.deployments.ScoringMetaNames.INPUT_DATA_REFERENCES

        if input_data_references in scoring_params:
            payload_ref = [input_ref for input_ref in scoring_params[input_data_references]]

            return self.run_job(payload=payload_ref, output_data_reference=scoring_params['output_data_reference'],
                                background_mode=background_mode)
        else:
            payload_df = DataFrame.from_dict(scoring_params['input_data'])

            return self.score(payload=payload_df, background_mode=background_mode)

    @BaseDeployment._project_to_space_to_project
    def delete(self, deployment_id: str = None) -> None:
        """
        Delete deployment on WML.

        Parameters
        ----------
        deployment_id: str, optional
            ID of the deployment to delete. If empty, current deployment will be deleted.

        Example
        -------
        >>> deployment = Batch(workspace=...)
        >>> # Delete current deployment
        >>> deployment.delete()
        >>> # Or delete a specific deployment
        >>> deployment.delete(deployment_id='...')
        """
        super().delete(deployment_id=deployment_id, deployment_type='batch')

    @BaseDeployment._project_to_space_to_project
    def list(self, limit=None) -> 'DataFrame':
        """
        List WML deployments.

        Parameters
        ----------
        limit: int, optional
            Set the limit of how many deployments to list. Default is None (all deployments should be fetched)

        Returns
        -------
        Pandas DataFrame with information about deployments.

        Example
        -------
        >>> deployment = Batch(workspace=...)
        >>> deployments_list = deployment.list()
        >>> print(deployments_list)
                             created_at  ...  status
            0  2020-03-06T10:50:49.401Z  ...   ready
            1  2020-03-06T13:16:09.789Z  ...   ready
            4  2020-03-11T14:46:36.035Z  ...  failed
            3  2020-03-11T14:49:55.052Z  ...  failed
            2  2020-03-11T15:13:53.708Z  ...   ready
        """
        return super().list(limit=limit, deployment_type='batch')

    @BaseDeployment._project_to_space_to_project
    def get(self, deployment_id: str) -> None:
        """
        Get WML deployment.


        Parameters
        ----------
        deployment_id: str, required
            ID of the deployment to work with.

        Returns
        -------
        WebService deployment object

        Example
        -------
        >>> deployment = Batch(workspace=...)
        >>> deployment.get(deployment_id="...")
        """
        super().get(deployment_id=deployment_id, deployment_type='batch')

    @BaseDeployment._project_to_space_to_project
    def get_job_params(self, scoring_job_id: str = None) -> Dict:
        """Get batch deployment job parameters.

        Parameters
        ----------
        scoring_job_id: str
           Id of scoring job.

        Returns
        -------
        Dictionary with parameters of the scoring job.
           """
        return self._target_workspace.wml_client.deployments.get_job_details(scoring_job_id)

    @BaseDeployment._project_to_space_to_project
    def get_job_status(self, scoring_job_id: str) -> Dict:
        """Get status of scoring job.
        Parameters
        ----------
        scoring_job_id: str
           Id of scoring job.

        Returns
        -------
        Dictionary with state of scoring job (one of: [completed, failed, starting, queued])
            and additional details if they exist.
        """
        return self._target_workspace.wml_client.deployments.get_job_status(scoring_job_id)

    @BaseDeployment._project_to_space_to_project
    def get_job_result(self, scoring_job_id: str) -> Union[Dict, 'DataFrame']:
        """Get batch deployment results of job with id `scoring_job_id`.

        Parameters
        ----------
        scoring_job_id: str
           Id of scoring job which results will be returned.

        Returns
        -------
        Dictionary with predictions for scoring job with inline input data or
            dictionary with data reference to output data if the scoring job has reference to input data.
            In case of incompleted or failed scoring None is returned.
        """
        scoring_params = self.get_job_params(scoring_job_id)['entity']['scoring']
        if scoring_params['status']['state'] == 'completed':
            if 'predictions' in scoring_params:
                return scoring_params['predictions']
            else:
                return self._download_results(scoring_params['output_data_reference'], scoring_job_id)
        else:
            raise MissingScoringResults(scoring_job_id, reason="Scoring is not completed.")

    @staticmethod
    def _download_results(data_connection: dict, scoring_job_id: str) -> Union[Dict, 'DataFrame']:
        """Download batch results."""

        if 's3' in str(data_connection):
            cos_client = init_cos_client(data_connection['connection'])

            # note: fetch all batch results file names per scoring_job_id
            cos_summary = cos_client.Bucket(data_connection['location']['bucket']).objects.filter(
                Prefix=data_connection['location']['path'])
            file_names = [file_name.key for file_name in cos_summary if scoring_job_id in file_name.key]
            # --- end note

            # TODO: this can be done simultaneously (multithreading / multiprocessing)
            # note: download all data parts and concatenate them into one output
            parts = []
            for file_name in file_names:
                file = cos_client.Object(data_connection['location']['bucket'], file_name).get()
                buffer = io.BytesIO(file['Body'].read())
                parts.append(try_load_dataset(buffer=buffer))

            data = concat(parts)
            # --- end note
            return data

        else:
            return data_connection

    def get_job_id(self, batch_scoring_details):
        """Get id from batch scoring details."""
        return self._target_workspace.wml_client.deployments.get_job_uid(batch_scoring_details)

    def list_jobs(self):
        """Returns pandas DataFrame with list of deployment jobs"""

        resources = self._target_workspace.wml_client.deployments.get_job_details()['resources']
        columns = [u'job id', u'state', u'creted', u'deployment id']
        values = []
        for scoring_details in resources:
            if 'scoring' in scoring_details['entity']:
                state = scoring_details['entity']['scoring']['status']['state']
                score_values = (scoring_details[u'metadata'][u'guid'], state,
                                scoring_details[u'metadata'][u'created_at'],
                                scoring_details['entity']['deployment']['id'])
                if self.id:
                    if self.id == scoring_details['entity']['deployment']['id']:
                        values.append(score_values)
                else:
                    values.append(score_values)

        return DataFrame(values, columns=columns)

    @BaseDeployment._project_to_space_to_project
    def _deploy(self,
                pipeline_model: 'Pipeline',
                deployment_name: str,
                meta_props: Dict,
                training_data: Union['DataFrame', 'ndarray'],
                training_target: Union['DataFrame', 'ndarray'],
                result_client=None) -> Dict:
        """
        Deploy model into WML.

        Parameters
        ----------
        pipeline_model: Union['Pipeline', str], required
            Model of the pipeline to deploy

        deployment_name: str, required
            Name of the deployment

        training_data: Union['pandas.DataFrame', 'numpy.ndarray'], required
            Training data for the model

        training_target: Union['pandas.DataFrame', 'numpy.ndarray'], required
            Target/label data for the model

        meta_props: dictionary, required
            Model meta properties.

        result_client: Tuple['DataConnection', 'resource'] required
            Tuple with Result DataConnection object and initialized COS client.
        """
        deployment_details = {}
        asset_uid = self._publish_model(pipeline_model=pipeline_model,
                                        meta_props=meta_props,
                                        training_data=training_data,
                                        training_target=training_target)

        self.asset_id = asset_uid

        deployment_props = {
            self._target_workspace.wml_client.deployments.ConfigurationMetaNames.NAME: deployment_name,
            self._target_workspace.wml_client.deployments.ConfigurationMetaNames.BATCH: {}
        }

        deployment_props[self._target_workspace.wml_client.deployments.ConfigurationMetaNames.ASSET] = {
            "id": asset_uid
        }

        deployment_props[
            self._target_workspace.wml_client.deployments.ConfigurationMetaNames.HYBRID_PIPELINE_HARDWARE_SPECS] = [
            {
                'node_runtime_id': 'autoai.kb',
                'hardware_spec': {
                    'name': 'M'
                }
            }
        ]

        print("Deploying model {} using V4 client.".format(asset_uid))
        try:
            deployment_details = self._target_workspace.wml_client.deployments.create(
                artifact_uid=asset_uid,
                meta_props=deployment_props)

            self.deployment_id = self._target_workspace.wml_client.deployments.get_uid(deployment_details)

        except WMLClientError as e:
            raise e

        return deployment_details
